
static const char *identifier = "Tue Jun 22 14:23:58 2021";

static const double T_r = 300.0;
static const double V_r = 4.36;
static const double a0 = -2055000.0;
static const double n = 7.0;
static const double v0 = 4.36;
static const double k00 = 1280000.0;
static const double k0p = 4.2;
static const double theta0 = 809.0;
static const double gamma0 = 0.99;
static const double q = 2.1;
static const double refS = 0.0;
static const double R = 8.314472;

#include <math.h>

#include <float.h>
#include <assert.h>

static double chebvalat(double x) {
    double c[17] = {
        2.707737068327440945 / 2.0, 0.340068135211091751, -0.12945150184440869e-01, 0.7963755380173816e-03,
        -0.546360009590824e-04, 0.39243019598805e-05, -0.2894032823539e-06, 0.217317613962e-07, -0.16542099950e-08,
        0.1272796189e-09, -0.987963460e-11, 0.7725074e-12, -0.607797e-13, 0.48076e-14, -0.3820e-15, 0.305e-16, -0.24e-17
    };
    double x2 = 2 * x;
    double c0 = c[17-2];
    double c1 = c[17-1];
    for (int i=3; i<18; i++) {
        double tmp = c0;
        c0 = c[17-i] - c1;
        c1 = tmp + c1 * x2;
    }
    return c0 + c1 * x;
}

static double Debye(double x) {
    //
    // returns D_3(x) = 3/x^3\int_0^t t^3/(e^t - 1) dt
    //
    
    double val_infinity = 19.4818182068004875;
    double sqrt_eps = sqrt(DBL_EPSILON);
    double log_eps = log(DBL_EPSILON);
    double xcut = -log_eps;

    //Check for negative x (was returning zero)
    assert(x >= 0.);

    if (x < (2.0*sqrt(2.0)*sqrt_eps)) return 1.0 - 3.0*x/8.0 + x*x/20.0;
    else if (x <= 4.0) {
        double t = x*x/8.0 - 1.0;
        double c = chebvalat(t);
        return c - 0.375*x;
    } else if (x < -(log(2.0)+log_eps)) {
        int nexp = (int)(floor(xcut / x));
        double ex = exp(-x);
        double xk = nexp * x;
        double rk = nexp;
        double sum = 0.0;
        for (int i=nexp; i>0; i--) {
            double xk_inv = 1.0/xk;
            sum *= ex;
            sum += (((6.0*xk_inv + 6.0)*xk_inv + 3.0)*xk_inv + 1.0)/rk;
            rk -= 1.0;
            xk -= x;
        }
        return val_infinity / (x * x * x) - 3.0 * sum * ex;
    } else if (x < xcut) {
        double x3 = x*x*x;
        double sum = 6.0 + 6.0*x + 3.0*x*x + x3;
        return (val_infinity - 3.0*sum*exp(-x))/x3;
    } else return ((val_infinity/x)/x)/x;
}

static double stixrude_a(double T, double V) {
    double result = 0.0;
    result += R*T*n*(3*log(1 - exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)) - Debye(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)) - R*T_r*n*(3*log(1 - exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T_r)) - Debye(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T_r)) + a0 + 9*k00*v0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(((1.0/2.0)*k0p - 2)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1.0/2.0);
    return result;
}

static double stixrude_dadt(double T, double V) {
    double result = 0.0;
    result += R*T*n*(theta0*(-3*T*Debye(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)/(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)) + 3/(exp(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T) - 1))*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/((T)*(T)) - 3*theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)*exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)/(((T)*(T))*(1 - exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)))) + R*n*(3*log(1 - exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)) - Debye(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T));
    return result;
}

static double stixrude_dadv(double T, double V) {
    double result = 0.0;
    result += R*T*n*(-theta0*(-3*T*Debye(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)/(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)) + 3/(exp(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T) - 1))*(-gamma0*pow(v0/V, 2.0/3.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*(6*gamma0 - 3*q - 2)/V - gamma0*pow(v0/V, 2.0/3.0)/V)/(T*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)) + 3*theta0*(-gamma0*pow(v0/V, 2.0/3.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*(6*gamma0 - 3*q - 2)/V - gamma0*pow(v0/V, 2.0/3.0)/V)*exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T)/(T*(1 - exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T))*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1))) - R*T_r*n*(-theta0*(-3*T_r*Debye(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T_r)/(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)) + 3/(exp(theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T_r) - 1))*(-gamma0*pow(v0/V, 2.0/3.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*(6*gamma0 - 3*q - 2)/V - gamma0*pow(v0/V, 2.0/3.0)/V)/(T_r*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)) + 3*theta0*(-gamma0*pow(v0/V, 2.0/3.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*(6*gamma0 - 3*q - 2)/V - gamma0*pow(v0/V, 2.0/3.0)/V)*exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T_r)/(T_r*(1 - exp(-theta0*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1)/T_r))*sqrt(3*gamma0*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))*(6*gamma0 - 3*q - 2) + 6*gamma0*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1))) - 3*k00*v0*pow(v0/V, 2.0/3.0)*((1.0/2.0)*k0p - 2)*(((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0))/V - 6*k00*v0*pow(v0/V, 2.0/3.0)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0)*(((1.0/2.0)*k0p - 2)*((1.0/2.0)*pow(v0/V, 2.0/3.0) - 1.0/2.0) + 1.0/2.0)/V;
    return result;
}

static double stixrude_d2adt2(double T, double V) {
    double result = 0.0;
    result += 3*R*n*theta0*(sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(-Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 3*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/T + theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)))) + 2*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T - 2*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1) + exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))/T + 2*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) - 1.0/4.0*theta0*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) - 1.0/4.0*theta0*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))))/T;
    return result;
}

static double stixrude_d2adtdv(double T, double V) {
    double result = 0.0;
    result += -3.0/2.0*R*gamma0*n*theta0*pow(v0/V, 2.0/3.0)*(-(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + (-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1) + exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(T*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1))) - (-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) + theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))))/(T*V);
    return result;
}

static double stixrude_d2adv2(double T, double V) {
    double result = 0.0;
    result += pow(v0/V, 2.0/3.0)*(-1.0/4.0*R*gamma0*n*theta0*(3*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0) - 3*gamma0*pow(v0/V, 2.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(-T*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + 3*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 2*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + 2*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 12*gamma0*theta0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 12*gamma0*theta0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) + (1.0/4.0)*R*gamma0*n*theta0*(3*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0) - 3*gamma0*pow(v0/V, 2.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(-T_r*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + 3*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 2*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + 2*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 12*gamma0*theta0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 12*gamma0*theta0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) + k00*v0*pow(v0/V, 2.0/3.0)*(k0p - 4)*(pow(v0/V, 2.0/3.0) - 1) + (1.0/2.0)*k00*v0*pow(v0/V, 2.0/3.0)*((k0p - 4)*(pow(v0/V, 2.0/3.0) - 1) + 2) + (5.0/8.0)*k00*v0*(k0p - 4)*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1)) + (5.0/4.0)*k00*v0*(pow(v0/V, 2.0/3.0) - 1)*((k0p - 4)*(pow(v0/V, 2.0/3.0) - 1) + 2))/((V)*(V));
    return result;
}

static double stixrude_d3adt3(double T, double V) {
    double result = 0.0;
    result += 3*R*n*theta0*(3*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(-Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 3*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/T + theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)))) + 4*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 3*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/T - theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)))) + (1.0/4.0)*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(-12*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 36*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/T + 4*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))) - ((theta0)*(theta0))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))) + 2*((theta0)*(theta0))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*exp(2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)))) - 6*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T + 6*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T + (3.0/4.0)*theta0*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) + (3.0/4.0)*theta0*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))) - ((theta0)*(theta0))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) - 3*((theta0)*(theta0))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))) - 2*((theta0)*(theta0))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)*exp(-3*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))))/((T)*(T));
    return result;
}

static double stixrude_d3adt2dv(double T, double V) {
    double result = 0.0;
    result += -3*R*gamma0*n*theta0*pow(v0/V, 2.0/3.0)*(-1.0/2.0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 3*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/T - theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + (1.0/2.0)*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(T*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 3*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 12*(-T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 1.0/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/T + theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + ((theta0)*(theta0))*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))) - 2*((theta0)*(theta0))*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)))) + (-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(T*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(T*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/T + (-(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(T*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1))) - (-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) + theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))))/T + (-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 2*theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) - 2*theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))) + (1.0/2.0)*((theta0)*(theta0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) + (3.0/2.0)*((theta0)*(theta0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))) + ((theta0)*(theta0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*exp(-3*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))))/(T*V);
    return result;
}

static double stixrude_d3adtdv2(double T, double V) {
    double result = 0.0;
    result += -1.0/4.0*R*gamma0*n*theta0*pow(v0/V, 2.0/3.0)*(3*gamma0*pow(v0/V, 2.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(-T*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)*(9*T*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) - 2*T*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 432*gamma0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 36*gamma0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) + 24*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) + 3*gamma0*theta0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 2*theta0*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 12*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) - 24*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) + 2*theta0*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))) + 2*theta0*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))) + 3*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 9*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 6*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-3*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/(T*((V)*(V)));
    return result;
}

static double stixrude_d3adv3(double T, double V) {
    double result = 0.0;
    result += (-1.0/24.0*R*gamma0*n*theta0*(-6*gamma0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)/pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0) + 3*gamma0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(9*T*gamma0*pow(v0/V, 2.0/3.0)*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) - 2*T*(2*pow(v0/V, 2.0/3.0)*(6*gamma0 - 3*q - 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 10)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 432*gamma0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 36*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2))*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 12*(-T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 1.0/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - 24*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - 3*gamma0*theta0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 2*theta0*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 12*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2))*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) - 24*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2))*exp(2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) - 12*gamma0*pow(v0/V, 4.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)/pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0) + 12*gamma0*pow(v0/V, 4.0/3.0)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*(T*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) - 6*gamma0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 12*gamma0*pow(v0/V, 4.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 40*pow(v0/V, 2.0/3.0)*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*(-3*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 4*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 8)/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + 40*pow(v0/V, 2.0/3.0)*(-3*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 4*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 8)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 27*((gamma0)*(gamma0))*((v0)*(v0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*(T*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(((V)*(V))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) - 18*((gamma0)*(gamma0))*((v0)*(v0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(T*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/(((V)*(V))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 27*((gamma0)*(gamma0))*((v0)*(v0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((V)*(V))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) + 72*gamma0*theta0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 72*gamma0*theta0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) - 432*((gamma0)*(gamma0))*theta0*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((V)*(V))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) - 432*((gamma0)*(gamma0))*theta0*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(T*((V)*(V))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) + 9*((gamma0)*(gamma0))*((theta0)*(theta0))*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((V)*(V))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 27*((gamma0)*(gamma0))*((theta0)*(theta0))*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((V)*(V))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 18*((gamma0)*(gamma0))*((theta0)*(theta0))*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(-3*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)/(((T)*(T))*((V)*(V))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T)))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0))) + (1.0/24.0)*R*gamma0*n*theta0*(-6*gamma0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)/pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0) + 3*gamma0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(9*T_r*gamma0*pow(v0/V, 2.0/3.0)*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) - 2*T_r*(2*pow(v0/V, 2.0/3.0)*(6*gamma0 - 3*q - 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 10)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 432*gamma0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T_r*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 36*gamma0*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2))*(T_r*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*pow((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 12*(-T_r*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 1.0/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/(3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - 24*(T_r*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - 3*gamma0*theta0*pow(v0/V, 2.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 2*theta0*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 12*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2))*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(((T_r)*(T_r))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) - 24*gamma0*((theta0)*(theta0))*pow(v0/V, 2.0/3.0)*(((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2))*exp(2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(((T_r)*(T_r))*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) - 12*gamma0*pow(v0/V, 4.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)/pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0) + 12*gamma0*pow(v0/V, 4.0/3.0)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*(T_r*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T_r*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) - 6*gamma0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) - 12*gamma0*pow(v0/V, 4.0/3.0)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 40*pow(v0/V, 2.0/3.0)*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*(-3*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 4*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 8)/sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1) + 40*pow(v0/V, 2.0/3.0)*(-3*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) - 4*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 8)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) + 27*((gamma0)*(gamma0))*((v0)*(v0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*(T_r*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/(((V)*(V))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) - 18*((gamma0)*(gamma0))*((v0)*(v0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*(T_r*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*Debye(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 12*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*(T_r*Debye(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)) - 1/(exp(theta0*sqrt((3.0/4.0)*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(6*gamma0 - 3*q - 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))/(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4) - theta0*((pow(v0/V, 2.0/3.0) - 1)*(6*gamma0 - 3*q - 2) + 2)*exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1)*(exp(theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r) - 1))*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)))/(((V)*(V))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 27*((gamma0)*(gamma0))*((v0)*(v0))*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(((V)*(V))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 5.0/2.0)) + 72*gamma0*theta0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) + 72*gamma0*theta0*pow(v0/V, 4.0/3.0)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(2*pow(v0/V, 2.0/3.0)*(-6*gamma0 + 3*q + 2) + 5*(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 10)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)))*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)) - 432*((gamma0)*(gamma0))*theta0*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((V)*(V))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) - 432*((gamma0)*(gamma0))*theta0*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*((pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) - 2)*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(T_r*((V)*(V))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)))*((-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4)*(-3*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 12*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 4))) + 9*((gamma0)*(gamma0))*((theta0)*(theta0))*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(((T_r)*(T_r))*((V)*(V))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 27*((gamma0)*(gamma0))*((theta0)*(theta0))*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(-2*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(((T_r)*(T_r))*((V)*(V))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0)) + 18*((gamma0)*(gamma0))*((theta0)*(theta0))*((v0)*(v0))*((-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2)*(-(pow(v0/V, 2.0/3.0) - 1)*(-6*gamma0 + 3*q + 2) + 2))*exp(-3*theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)/(((T_r)*(T_r))*((V)*(V))*((1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r))*(1 - exp(-theta0*sqrt(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1)/T_r)))*pow(-3.0/4.0*gamma0*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1))*(-6*gamma0 + 3*q + 2) + 3*gamma0*(pow(v0/V, 2.0/3.0) - 1) + 1, 3.0/2.0))) - 5*k00*v0*pow(v0/V, 4.0/3.0)*(k0p - 4)*(pow(v0/V, 2.0/3.0) - 1) - 5.0/2.0*k00*v0*pow(v0/V, 4.0/3.0)*((k0p - 4)*(pow(v0/V, 2.0/3.0) - 1) + 2) - 5.0/3.0*k00*v0*pow(v0/V, 2.0/3.0)*(k0p - 4)*((pow(v0/V, 2.0/3.0) - 1)*(pow(v0/V, 2.0/3.0) - 1)) - 10.0/3.0*k00*v0*pow(v0/V, 2.0/3.0)*(pow(v0/V, 2.0/3.0) - 1)*((k0p - 4)*(pow(v0/V, 2.0/3.0) - 1) + 2) - k00*((v0)*(v0)*(v0))*(k0p - 4)/((V)*(V)))/((V)*(V)*(V));
    return result;
}


static double V = 2.0;
static void stixrude_solve_V(double T, double P) {
    static double Told = 0.0;
    static double Pold = 0.0;
    // Update if *either* T or P changes (was only if both changed)
    // if ((T != Told) && (P != Pold)) {
    if ((T != Told) || (P != Pold)) {
        Told = T;
        Pold = P;
        double f = 0.0;
        int iter = 0;
        do {
            f = -stixrude_dadv(T, V) - P;
            double df = -stixrude_d2adv2(T, V);
            if (df == 0.0) break;
            V -= f/df;
            if (V <= 0.0) V = 0.001;
            else if (V >= 4.0*2.0) V = 4.0*2.0;
            iter++;
        } while ((fabs(f) > 0.001) && (iter < 200));
    }
}

static double stixrude_g(double T, double P) {
    stixrude_solve_V(T, P);
    double A = stixrude_a(T, V);
    return A + P*V;
}

static double stixrude_dgdt(double T, double P) {
    stixrude_solve_V(T, P);
    double dAdT = stixrude_dadt(T, V);
    return dAdT;
}

static double stixrude_dgdp(double T, double P) {
    stixrude_solve_V(T, P);
    return V;
}

static double stixrude_d2gdt2(double T, double P) {
    stixrude_solve_V(T, P);
    double d2AdT2 = stixrude_d2adt2(T, V);
    double d2AdTdV = stixrude_d2adtdv(T, V);
    double d2AdV2 = stixrude_d2adv2(T, V);
    return d2AdT2 - d2AdTdV*d2AdTdV/d2AdV2;
}

static double stixrude_d2gdtdp(double T, double P) {
    stixrude_solve_V(T, P);
    double d2AdTdV = stixrude_d2adtdv(T, V);
    double d2AdV2 = stixrude_d2adv2(T, V);
    return - d2AdTdV/d2AdV2;
}

static double stixrude_d2gdp2(double T, double P) {
    stixrude_solve_V(T, P);
    double d2AdV2 = stixrude_d2adv2(T, V);
    return - 1.0/d2AdV2;
}

static double stixrude_d3gdt3(double T, double P) {
    stixrude_solve_V(T, P);
    double d3AdT3 = stixrude_d3adt3(T, V);
    double d3AdT2dV = stixrude_d3adt2dv(T, V);
    double d3AdTdV2 = stixrude_d3adtdv2(T, V);
    double d3AdV3 = stixrude_d3adv3(T,V);
    double d2AdTdV = stixrude_d2adtdv(T, V);
    double d2AdV2 = stixrude_d2adv2(T, V);
    double dVdT = -d2AdTdV/d2AdV2;
    double d2VdT2 = (-d3AdT2dV - 2.0*d3AdTdV2*dVdT - d3AdV3*dVdT*dVdT)/d2AdV2;
    return d3AdT3 + 2.0*d3AdT2dV*dVdT + d3AdTdV2*dVdT*dVdT + d2AdTdV*d2VdT2;
}

static double stixrude_d3gdt2dp(double T, double P) {
    stixrude_solve_V(T, P);
    double d3AdT2dV = stixrude_d3adt2dv(T, V);
    double d3AdTdV2 = stixrude_d3adtdv2(T, V);
    double d3AdV3 = stixrude_d3adv3(T,V);
    double d2AdTdV = stixrude_d2adtdv(T, V);
    double d2AdV2 = stixrude_d2adv2(T, V);
    double dVdT = -d2AdTdV/d2AdV2;
    double dVdP = -1.0/d2AdV2;
    double d2VdTdP = (-d3AdTdV2*dVdP - d3AdV3*dVdT*dVdP)/d2AdV2;
    return d3AdT2dV*dVdP + d3AdTdV2*dVdT*dVdP + d2AdTdV*d2VdTdP;
}

static double stixrude_d3gdtdp2(double T, double P) {
    stixrude_solve_V(T, P);
    double d3AdTdV2 = stixrude_d3adtdv2(T, V);
    double d3AdV3 = stixrude_d3adv3(T,V);
    double d2AdTdV = stixrude_d2adtdv(T, V);
    double d2AdV2 = stixrude_d2adv2(T, V);
    double dVdT = -d2AdTdV/d2AdV2;
    return (d3AdTdV2 + d3AdV3*dVdT)/d2AdV2/d2AdV2;
}

static double stixrude_d3gdp3(double T, double P) {
    stixrude_solve_V(T, P);
    double d3AdV3 = stixrude_d3adv3(T,V);
    double d2AdV2 = stixrude_d2adv2(T, V);
    double dVdP = -1.0/d2AdV2;
    return d3AdV3*dVdP/d2AdV2/d2AdV2;
}

static double stixrude_s(double T, double P) {
    double result = -stixrude_dgdt(T, P);
    return result;
}

static double stixrude_v(double T, double P) {
    double result = stixrude_dgdp(T, P);
    return result;
}

static double stixrude_cv(double T, double P) {
    double result = -T*stixrude_d2gdt2(T, P);
    double dvdt = stixrude_d2gdtdp(T, P);
    double dvdp = stixrude_d2gdp2(T, P);
    result += T*dvdt*dvdt/dvdp;
    return result;
}

static double stixrude_cp(double T, double P) {
    double result = -T*stixrude_d2gdt2(T, P);
    return result;
}

static double stixrude_dcpdt(double T, double P) {
    double result = -T*stixrude_d3gdt3(T, P) - stixrude_d2gdt2(T, P);
    return result;
}

static double stixrude_alpha(double T, double P) {
    double result = stixrude_d2gdtdp(T, P)/stixrude_dgdp(T, P);
    return result;
}

static double stixrude_beta(double T, double P) {
    double result = -stixrude_d2gdp2(T, P)/stixrude_dgdp(T, P);
    return result;
}

static double stixrude_K(double T, double P) {
    double result = -stixrude_dgdp(T, P)/stixrude_d2gdp2(T, P);
    return result;
}

static double stixrude_Kp(double T, double P) {
    double result = stixrude_dgdp(T, P);
    result *= stixrude_d3gdp3(T, P);
    result /= pow(stixrude_d2gdp2(T, P), 2.0);
    return result - 1.0;
}


const char *Forsterite_stixrude_identifier(void) {
    return identifier;
}

const char *Forsterite_stixrude_name(void) {
    return "Forsterite";
}

const char *Forsterite_stixrude_formula(void) {
    return "Mg2SiO4";
}

const double Forsterite_stixrude_mw(void) {
    return 140.6931;
}

static const double elmformula[106] = {
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,4.0,0.0,0.0,0.0,
        2.0,0.0,1.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0,0.0,0.0,
        0.0,0.0,0.0,0.0
    };

const double *Forsterite_stixrude_elements(void) {
    return elmformula;
}

double Forsterite_stixrude_g(double T, double P) {
    return stixrude_g(T, P);
}

double Forsterite_stixrude_dgdt(double T, double P) {
    return stixrude_dgdt(T, P);
}

double Forsterite_stixrude_dgdp(double T, double P) {
    return stixrude_dgdp(T, P);
}

double Forsterite_stixrude_d2gdt2(double T, double P) {
    return stixrude_d2gdt2(T, P);
}

double Forsterite_stixrude_d2gdtdp(double T, double P) {
    return stixrude_d2gdtdp(T, P);
}

double Forsterite_stixrude_d2gdp2(double T, double P) {
    return stixrude_d2gdp2(T, P);
}

double Forsterite_stixrude_d3gdt3(double T, double P) {
    return stixrude_d3gdt3(T, P);
}

double Forsterite_stixrude_d3gdt2dp(double T, double P) {
    return stixrude_d3gdt2dp(T, P);
}

double Forsterite_stixrude_d3gdtdp2(double T, double P) {
    return stixrude_d3gdtdp2(T, P);
}

double Forsterite_stixrude_d3gdp3(double T, double P) {
    return stixrude_d3gdp3(T, P);
}

double Forsterite_stixrude_s(double T, double P) {
    return stixrude_s(T, P);
}

double Forsterite_stixrude_v(double T, double P) {
    return stixrude_v(T, P);
}

double Forsterite_stixrude_cv(double T, double P) {
    return stixrude_cv(T, P);
}

double Forsterite_stixrude_cp(double T, double P) {
    return stixrude_cp(T, P);
}

double Forsterite_stixrude_dcpdt(double T, double P) {
    return stixrude_dcpdt(T, P);
}

double Forsterite_stixrude_alpha(double T, double P) {
    return stixrude_alpha(T, P);
}

double Forsterite_stixrude_beta(double T, double P) {
    return stixrude_beta(T, P);
}

double Forsterite_stixrude_K(double T, double P) {
    return stixrude_K(T, P);
}

double Forsterite_stixrude_Kp(double T, double P) {
    return stixrude_Kp(T, P);
}

